/* 	$Id: ga.c,v 1.9 2007/08/31 13:20:05 cgm22 Exp cgm22 $	 */

#include <assert.h>
#include <errno.h>
#include <malloc.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#include <sys/time.h>
#include <sys/wait.h>

#include "unitor.h"

#ifdef _MICROBIAL_GA_

void
ga(char **pop, float fitness[], int lineage[], char timestamp[][42], char *genotype, float (*lifecycle)(char *, char *))
{
	int a, b, w, l, i, j, k, gen;


#ifdef CONSOLEDUMP
	printf("Initialising population & calculating initial fitness values \n");
#endif

	for (i = 0;i < POPULATION;i++) {
		for (j = 0;j < GENELENGTH;j++) {
		  setbit(pop[i], j, (drand48() < 0.5));
		}

		/* if (drand48() < 0.5) memcpy(pop[i], genotype, GENELENGTHB); */

		fitness[i] = lifecycle(pop[i], timestamp[i]);
		lineage[i] = 0;

#ifdef CONSOLEDUMP
		printf("timestamp: %s, individual: %d, fitness = %f\n", timestamp[i], i, fitness[i]);
#endif
		fprintf(fd_logfile, "%s\t %d\t %f\t %d\n", timestamp[i], i, fitness[i], 0);
	}

#ifdef CONSOLEDUMP
	printf("Starting Microbial GA \n");
#endif
	
	for (gen = 0;gen < MAXTOURNAMENTS;gen++) {

#ifdef CONSOLEDUMP
		printf("Tournament: %d \n", gen);
#endif

		/* The following algorithm is based on 
		 * Dr. Inman Harvey's EaSy MSc. Teaching slides.
		 */

#define REC 0.5
#define MUT (1.0/(GENELENGTH))

		/* Pick two out of the POPULATION */

		a = POPULATION * drand48();

		do {
			b = POPULATION * drand48();
		} while (a == b); /* make sure a and b are different */


		/*  fitness function gives higher number for better fits. */

		if (fitness[a] > fitness[b]) {
			w = a;
			l = b;
		}
		else {
			w = b;
			l = a;
		}

		/* Tournament Loop */
		for (i = 0;i < GENELENGTH;i++) {
			if (drand48() < REC) { /* cross with probability REC
						* */
				setbit(pop[l], i, getbit(pop[w], i));
				lineage[l] = gen;
			}

			if (drand48() < MUT) { /* mutate with probability MUT
						* */

				/* flip bit */
				setbit(pop[l], i, !getbit(pop[l], i));
				lineage[l] = gen;
			}

		} /* end tournament loop */

		if (lineage[l] == gen) fitness[l] = lifecycle(pop[l], timestamp[l]);

#ifdef CONSOLEDUMP
		printf("timestamp: %s, individual: %d, fitness = %f\n", timestamp[l], l, fitness[l]);
#endif
		fprintf(fd_logfile, "%s\t %d\t %f\t %d\n", timestamp[l], l, fitness[l], gen);
	}


	for (i = 0;i < POPULATION;i++) {
#ifdef DEBUG
		printf("Index: %d\tTimestamp: %s\tFitness: %f%% \n", i, timestamp[i], fitness[i]);
#endif
	}


}

#endif /* _MICROBIAL_GA_ */


/* XXX: The gray code routines are from open domain code posted by 
 *	Jerry Avins <jya@ieee.org> to usenet newsgroup comp.dsp
 *	See: http://www.dspguru.com/comp.dsp/tricks/alg/grayconv.htm
 *	Suitably adjusted for allele_t
 */

/*	The purpose of this function is to convert an unsigned
 *	binary number to reflected binary Gray code.
 */

allele_t binaryToGray8bit(allele_t num)
{
  return (num>>1) ^ num;
}

/* A tricky Trick: for up to 2^n bits, you can convert Gray to binary by */
/*   performing (2^n) - 1 binary-to Gray conversions. All you need is the */
/* function above and a 'for' loop. */

  /*
        The purpose of this function is to convert a reflected binary
        Gray code number to a binary number.
  */
allele_t grayToBinary8bit(allele_t num)
{
  allele_t temp = num ^ (num>>8);
  temp ^= (temp>>4);
  temp ^= (temp>>2);
  temp ^= (temp>>1);
  return temp;
}

/* Assumes Little Endian */
allele_t 
getgeneplusplus(allele_t **geneptr, int *nibble, int nibblesize)
{
	allele_t genelsb, genemsb;
	allele_t genemask;

	const int nibblemax = sizeof(allele_t) * 8;

	assert(nibblesize <= nibblemax);

	genemask = ((1 << nibblesize) - 1); 

	if ((*nibble) >= nibblemax) {
		(*geneptr) += ((*nibble) / nibblemax);
		(*nibble) = ((*nibble) % nibblemax);
	}


	genelsb = **geneptr;
	genelsb >>= (*nibble);
	genelsb &= genemask;
	
	if ( ((*nibble) + nibblesize) > nibblemax) {
		genemsb = *(*geneptr + 1);
		genemsb <<= (nibblemax - (*nibble));
		genemsb &= genemask;
		genelsb |= genemsb;
	}

	(*nibble) += nibblesize;
	return genelsb;
}	

int
getbit(char *string, int bitpos)
{
	return ((string[bitpos >> 3] >> (bitpos & 0x7)) & 0x1);
}

int
setbit(char *string, int bitpos, int value)
{

#if 0
	printf("bitpos: %d, bitpos(bytes): %d, bitpos(bit): %d, val(bitpos(bit)): %d\n",
	       bitpos, bitpos >> 3, bitpos & 0x7, (1 << (bitpos & 0x7)));

#endif
	if(value != 0) {
		string[bitpos >> 3] |= (1 << (bitpos & 0x7));
	}

	else {
		string[bitpos >> 3] &= ~(1 << (bitpos & 0x7));
	}
}


int voutnodedecode(allele_t gene)
{
	gene = grayToBinary8bit(gene);
	
	return (gene % (MAXNODES - 1) + 2);
}


/* Creates an individual Netlist file with a unique filename created
 * from the timestamp at the moment of creation.
 * Returns fitness of the individual.
 */

float
lifecycle(char *genestring, char *timestamp)
{

	char *netlist;
	int numcomps = 0;	/* Number of components in the netlist */

	struct samplesim simdata[SAMPLEMAX];
	int sample = 0;
	float fitness = 0;

	FILE *fd_cirfile, *fd_simdump;
	char cirfilename[52];	/* = TMPDIR"$timestamp.cir"; */
	char simdumpfname[56];	/* /tmp/cir-$timestamp.dmp */


	/* Allocate memory for the netlist */
	netlist = malloc(NETLISTSZ);


	gettimestamp(timestamp); /* Writes timestamp in the string
				  * timestamp.
				  */

	/* Fetch the netlist decoded from the genestring */
	if(numcomps = 0, (gen_netlist(timestamp, genestring, netlist, &numcomps, NETLISTSZ) == -1)) {
		/* If it is zero, it is an unviable netlist 
		 * See: gen_netlist() for details
		 * Bailout after gettimestamp() 
		 */

		free(netlist);
		return 0;
	}

	sprintf(cirfilename, TMPDIR"%s.cir", timestamp);

	fd_cirfile = fopen(cirfilename, "w");

	if (errno) {
		err(errno, "%s ", cirfilename);
		return -1;
	}

	fprintf(fd_cirfile, "%s", netlist);

	/* fflush(fd_cirfile); XXX: This is a fence before
	 * run_simulation. if either fflush() or fclose() is not
	 * called before run_simulation(), there's no guarantee the
	 * generated spicefile has been written to disk.
	 */
	/* fflush(fd_cirfile); */
	fclose(fd_cirfile);

	sprintf(simdumpfname, TMPDIR"cir-%s.dmp", timestamp);

	free(netlist);

	sprintf(simdumpfname, TMPDIR"cir-%s.dmp", timestamp);

	run_simulation(cirfilename, simdumpfname);

	/* Open the spice-run data file */
	fd_simdump = fopen(simdumpfname, "r");

	/* Quit if we can't open it. */
	if (errno) {
		printf("pid = %d\n", getpid());
		err(errno, "%s ", simdumpfname);
		return -1;
	}


	/* ngspice splits the .PRINT data into analysis specific 
	 * layouts. run_simulation() wrapper magic takes care of 
	 * lining them one after the other and removing "rubbish", 
	 * however, its best to read in and double check that the
	 * index numbers do indeed match. Hence the assert()s
	 * below.
	 */

	while(
		(EOF != fscanf(fd_simdump, "%d %f, %*f %f",
			       &simdata[sample].num, &simdata[sample].freq,
			       &simdata[sample].VM))

		) {

#if 0
		printf("%d %e %e\n", simdata[sample].num, simdata[sample].freq,
		       simdata[sample].VM);
#endif

		sample++;
		if (sample >= SAMPLEMAX) { /* Something is wrong... netlist is not working */
			printf("WARNING: netlist %s discarded. \n", timestamp);
			fclose(fd_simdump);
			return 0;
		}

	}

	if((simdata[sample - 1].num > SAMPLEMAX) || (simdata[sample - 1].num < 10 /* XXX: Arbitrary. */)) { /* Something is wrong... netlist is not working */
		printf("WARNING: netlist %s discarded. only %d samples\n", timestamp, simdata[sample - 1].num);
		fclose(fd_simdump);
		return 0;
	}


	fitness = fitness_correlation(simdata,sample);

	/* Case 3: Conditional. either multiply or add */
	if (fitness > 4.31) {

	/* Case 1: Add fitness components */
	/* fitness +=  (float) numcomps / MAXDEVS;  */

	/* Case 2: Multiply fitness components */
	fitness *=  (float) numcomps / MAXDEVS; 
	
	}
	
	/* Encourage number of components to be minimum. Currently
	 * this is meant to put pressure on the number of components
	 * per unitor.
	 * XXX: Find a way to put selective pressure on encouraging
	 * unitors to "coalesce" into real components. 
	 */

	fclose(fd_simdump);

	return fitness;
}

struct samplesim tgt[] = MBP3H8H8K11K;

float fitness_correlation(struct samplesim *psimdata, int samplecount)
{

	int i = 0, j = 0;

	unsigned int weight = 0;

	float fitness = 0;
	float m, vmx, vmt, vm1, vm2, vmdiff, sx, s1, s2; /* XXX: Optimise me please. */

	/*
	 * The loop below is a bit tricky. Basically what we're doing
	 * is to iterate along the sampled frequencies and find the
	 * target frequencies which enclose it. eg: if the sample
	 * frequency of sample 5 == 2000Hz, then find the nearest
	 * pair, say tgt[10], tgt[11] such that tgt[10] == 1500Hz, say,
	 * and tgt[11].freq == 2001 Hz. The target VM value is then
	 * scaled linearly to match the frequency of the sample. This
	 * makes the absolute VM comparison between the sample and the
	 * target justifiable.
	 */

	for (i = 0; i < samplecount; i++) {

		while (((j + 2) < (sizeof(tgt)/sizeof(tgt[0])))) {
			if (tgt[j + 1].freq <= psimdata[i].freq)
				j++;
			else
				break;
		}

		s1 = tgt[j].freq;
		s2 = tgt[j + 1].freq;

		vm1 = tgt[j].VM;
		vm2 = tgt[j + 1].VM;
		
		m = (vm2 - vm1)/(s2 - s1); /* slope to scale VM */
		
		sx = psimdata[i].freq;
		vmx = psimdata[i].VM;

		vmt = vm1 + m * (sx - s1);

		m = (tgt[j + 1].num - tgt[j].num)/(s2 - s1);	/* slope to scale the weight. */

		weight = tgt[j].num + m * (sx - s1);

		vmdiff = fabs(vmt - vmx);	/* Correlate between actual frequency
						 * and required frequency 
						 */

		if (isnan(vmdiff) || isinf(vmdiff)) /* Reject bad Voltage Mag values */
			return 0;

		/* ceiling at 1 */

		if (vmdiff > 1) continue; /* Ignore deviation vmdiff > 1, ie; (1 - vmdiff) == 0; ie; fitness += 0; line below is unnecessary */

		fitness += ((1 - vmdiff) * weight)/(-1U); 

	}
	return fitness * 100 / samplecount;

}

/* Writes the timestamp string into char *timestamp */
int gettimestamp(char *timestamp)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return	sprintf(timestamp, "%d-%d", tv.tv_sec, tv.tv_usec);
}

/* returns a string with filename the results are stored in */
void
run_simulation(char *cirfilename, char *simdumpfname)
{
	
	pid_t cpid, w;

	int childstatus;

	char spicecmdline[512];

	/* XXX: The following command line is extreme hackery to get
	 * around ngspice's screwed up .PRINT format. It is fragile, and
	 * needs to be used with caution. Double checking /tmp/*.cir
	 * for layout muckup is a first step with debug.
	 */

	sprintf(spicecmdline, 
		"(" SPICEBIN " -b %s | grep -e '^[[:digit:]]*[[:space:]]*[[:digit:]]' | sort -k 1,1 -g -s > %s) >& /dev/null",
		cirfilename,
		simdumpfname);

#ifdef DEBUG
	printf("cmdline: %s\n",
	       spicecmdline);
#endif	

	switch (cpid = fork()) {
	case -1:
		printf("error vforking!");
		if (errno) {
			err(errno);
		}

		exit(1); /* XXX: redundant ? */
		break;

	case 0:

#ifdef DEBUG
		printf("I'm the child. \n");
#endif


		if ( execlp("sh", "sh", "-c", spicecmdline, (char *) NULL) < 0) {
			if (errno) {
				printf("exec returned !!! to pid = %d\n", getpid());
				err(errno, "%s ", simdumpfname);
			}

		};
		_exit(0);
		break;

	default:
#ifdef DEBUG
		printf("I'm the parent. waiting on child pid = %d \n", cpid);
#endif

		/* The following waitpid() sequence Plagiarised from man waitpid() */

               do {
                   w = waitpid(cpid, &childstatus, WUNTRACED);
                   if (w == -1) { perror("waitpid"); exit(EXIT_FAILURE); }

		   if (WIFSIGNALED(childstatus)) {
                       printf("killed by signal %d\n", WTERMSIG(childstatus));
                   } else if (WIFSTOPPED(childstatus)) {
                       printf("stopped by signal %d\n", WSTOPSIG(childstatus));
                   }
               } while (!WIFEXITED(childstatus) && !WIFSIGNALED(childstatus));


	}

}

void
catch_INT(int sig_num)
{
	signal(SIGINT, catch_INT);
	printf("Caught ctrl-C. Log written to %s \n", logfilename);
	exit(0);
}

