/* 	$Id: rlc.c,v 1.3 2007/08/31 13:20:24 cgm22 Exp cgm22 $	 */

/* 	Derived from evolvebistableK.c,v 1.2 2007/04/23 01:07:59 cgm22 Exp cgm22 $	 */

/* 	Derived from evolvebistable.c,v 1.11 2007/04/21 22:09:34 cgm22 Exp cgm22 $	 */

/* 	Derived from AdSys evolvegate.c,v 1.4 2007/04/10 21:46:28 cherry Exp cherry $	 */

/* 	Derived from Alife evolvegate.c,v 1.9 2007/01/11 15:42:40 cherry Exp cherry $	 */


#include <assert.h>
#include <errno.h>
#include <malloc.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#include <sys/time.h>
#include <sys/wait.h>

#include "unitor.h"

/* 
 * The LUT has 48 entries. ( 3 tables of 16 entries each ). 
 * The index into the LUT is encoded in 6 bits.  
 * genedecode distributes the rest of the bits with even probability  
 * ( barring round off errors ).
 */

#define genedecode(gene) ((grayToBinary8bit(gene) * 48) / 64) 


int gen_device(int, int, int, char *, int *);
int gen_netlist(char *, char *, char *, int *, int);

float dpool[] = { /* Resistor */
		  LUTSHORTRES, 
		  10, 
		  20,
		  40,
		  80,
		  160,
		  320,
		  640,
		  2000, 
		  8000,
		  16000,
		  32000,
		  64000,
		  128000,
		  256000,
		  LUTSKIPRES,
		  /* Capacitor */
		  LUTSKIPCAP,
		  5.1e-8,
		  1.1e-7,
		  5.1e-7,
		  1.1e-6,
		  5.1e-6,
		  1.1e-5,
		  5.1e-5,
		  1.1e-4,
		  5.1e-4,
		  1.1e-3,
		  5.1e-3,
		  1.1e-2,
		  5.1e-2,
		  1.1e-1,
		  LUTSHORTCAP,
		  /* Inductor */
		  LUTSHORTIND,
		  1.1e-2,
		  2.1e-2,
		  4.1e-2,
		  8.1e-2,
		  1.1e-1,
		  2.1e-1,
		  4.1e-1,
		  8.1e-1,
		  1.1,
		  2.1,
		  4.1,
		  8.1,
		  10.1,
		  12.1,
		  LUTSKIPIND 
};


void catch_INT(int);

char logfilename[255];

FILE *fd_logfile;

int MAXDEVSLOG2 = 0; /* log2(MAXDEVS): initialised in main(). see below */
int MAXNODESLOG2 = 0; /* log2(MAXNODES): initialised in main(). see below */
int GENELENGTH = 0;  /* In bits. 3 elements per unitor. 1 voutnode extra.*/
int GENELENGTHB = 0; /* Fit genelength in bytes. */

main()
{

	int i;
	struct timeval tv;


	/* every individual in the population has the following
	 * properties: fitness, the timestamp of the instant it was
	 * born, the lineage stamp, or the generation number it was
	 * born. 
	 */

	float fitness[POPULATION];

	/* Assuming 64bits for the variable holding the timestamp in
	 * usec, the largest 64bit decimal number has 21digits. For
	 * two numbers: 42 digits. 
	 */

	char timestamp[POPULATION][42];
	int lineage[POPULATION];

	char *pop[POPULATION];

 	char *genotype = malloc(GENELENGTHB);
	allele_t allele;

	/* Install signal handlers */
	signal(SIGINT, catch_INT);

	/* Initialise MAXDEVSLOG2 */
	int maxdevs = MAXDEVS;
	MAXDEVSLOG2 = 0;
	do MAXDEVSLOG2++; while(maxdevs >>= 1);
	printf("MAXDEVS: %d, MAXDEVSLOG2: %d\n", MAXDEVS, MAXDEVSLOG2);

	/* Initialise MAXNODESLOG2 */
	int maxnodes = MAXNODES;
	MAXNODESLOG2 = 0;
	do MAXNODESLOG2++; while(maxnodes >>= 1);
	printf("MAXNODES: %d, MAXNODESLOG2: %d\n", MAXNODES, MAXNODESLOG2);

	GENELENGTH  = 6 * MAXDEVS + MAXDEVSLOG2;	/* In bits. 1 element per device. 1 voutnode extra.*/
	GENELENGTHB = (roundup(GENELENGTH, 8) / 8);	/* Fit genelength in bytes. */


	printf("GENELENGTHB = %d\n", GENELENGTHB);
	/* Initialise pop array of pointers */

	pop[0] = malloc(POPULATION * GENELENGTHB);

	for(i = 1;i < POPULATION;i++)
		pop[i] = pop[i - 1] + GENELENGTHB;
	
	sprintf(logfilename, "%s-%d.log", LOGFILEPREFIX, getpid());

	/* Open the log file for dumping log data */
	fd_logfile = fopen(logfilename, "w");

	/* Setup random number generator */

	gettimeofday(&tv, NULL);
	srand48(tv.tv_usec);

	/* initialise fitness array to 0 */
	
	bzero(fitness, POPULATION * sizeof(typeof(float)));
	bzero(lineage, POPULATION * sizeof(typeof(int)));

	ga(pop, fitness, lineage, timestamp, genotype, lifecycle);

	free(pop[0]);
	free(genotype);

	printf("Writing log to %s\n", logfilename);
	fclose(fd_logfile);

}

/* Writes netlist string into the given buffer *netlist 
 * of size strz
 * Returns the number of bytes written _excluding_ trailing '\0'.
 * or -1 on error
 */

int
gen_netlist(char *timestamp, char *genestring, char *netlist, int *numcompsp, int strsz)
{

	char *tmpp = netlist;
	int unumcomps = 0; /* Number of components per unitor */

	/* Increments geneptr if necessary. 
	 * Please don't touch nibble.  
	 */

	int nibble = 0;			/* A bit "pointer" */
	allele_t gene;
	allele_t *geneptr = genestring;	/* The geneptr points to the
					 * genestring and moves along it as it
					 * is processed by various
					 * "morphogenerators" 
					 */

	unsigned int lutp;	/* Index into LUT */


	unsigned int voutnode;

	int bytes;

	int i, j;

	assert(*numcompsp == 0);

	if (strsz < NETLISTSZ) return -1;

	/*****      Generate Circuit Netlist *****/

	netlist += sprintf(netlist, "%s%s%s", "*<---Individual number-*", timestamp, " *--->\n\n");
	/* Input at Node 1 XXX: Fixme. Make this more flexible. (evolvable ?) */
	netlist += sprintf(netlist, "%s",
			   "VS	1	0	AC	1	SIN(0	1	2KHZ)\n");


	/* Fetch the Vout Node */
	gene = getgeneplusplus(&geneptr, &nibble, MAXNODESLOG2); 
	voutnode = voutnodedecode(gene);

	if (!(voutnode <= MAXNODES)) {printf("voutnode: %d\n", voutnode); exit(0);}
	assert(voutnode <= MAXNODES);

	printf("voutnode: %d\n", voutnode); 

	if (voutnode == 0 || voutnode == 1) /* Don't need ground or Vin */
		return -1; 
			
	for (i = 0; i < MAXNODES; i++) {
		for (j = i + 1;j < MAXNODES; j++) {
			
			unumcomps = 0;
			
			/* Decode Device */
			gene = getgeneplusplus(&geneptr, &nibble, 6); 
			lutp = genedecode(gene);

			if (!((dpool[lutp] == LUTSKIPRES) || (dpool[lutp] == LUTSKIPCAP) || (dpool[lutp] == LUTSKIPIND))) {
				
				/* Write out component and component behaviour */
				netlist += sprintf(netlist, "X%d%d %d %d U%d%d\n", i, j, i, j, i, j);
				netlist += gen_device(lutp, i, j, netlist, &unumcomps);
				*numcompsp += unumcomps; /* Accumulate numcomps in the unitor */
			} else {
				/* Encourage parsimony ? */
				
				*numcompsp += 100;
			}
		}
	}

	netlist += sprintf(netlist, "%s", ".OPTIONS NOPAGE\n");
	netlist += sprintf(netlist, "%s", ".AC 	DEC 	5 10 1MEG\n");

	netlist += sprintf(netlist, "%s%d%s", ".PRINT  AC VM(", voutnode, ")\n"); 

	netlist += sprintf(netlist, "%s", ".END\n");
	return (netlist - tmpp);
}



/* Returns number of bytes sprintf()ed _excluding_ trailing '\0'. */
int gen_device(int lutp, int n1, int n2, char *buf, int *numcomps)
{

	float dev;
	char *tmpp = buf;

	dev = dpool[lutp];

	if ((dev == LUTSHORTRES) || (dev == LUTSHORTCAP) || (dev == LUTSHORTIND)) {
		
		/* Device is a short circuit ( 5 cases out of 27 )*/
		buf += sprintf(buf, ".SUBCKT U%d%d 1 3\n", n1, n2);
		buf += sprintf(buf, "R%d%d 1 3 %u\n", n1, n2, 0);
		buf += sprintf(buf, "%s", ".ENDS\n");
		*numcomps = 0;	/* short circuit == resistor of 0 ohms 
				 * XXX: Could also be capacitor of inf. cap. 
				 */
		
		return(buf - tmpp);	/* Number of bytes sprintf()ed */
	}


			
	if (lutp > 31) {
		/* Device is an inductor */
		buf += sprintf(buf, ".SUBCKT U%d%d 1 3\n", n1, n2);
		buf += sprintf(buf, "L%d%d 1 3 %e\n", n1, n2, dev);
		buf += sprintf(buf, "%s", ".ENDS\n");
		*numcomps = 100;	/* Inductor */
		return(buf - tmpp);	/* Number of bytes sprintf()ed */
	}

	/* else */

	if (lutp > 15) {
		/* Unitor is a capacitor */
		buf += sprintf(buf, ".SUBCKT U%d%d 1 3\n", n1, n2);
		buf += sprintf(buf, "C%d%d 1 3 %e\n", n1, n2, dev);
		buf += sprintf(buf, "%s", ".ENDS\n");
		*numcomps = 100;	/* Capacitor */
		return(buf - tmpp);	/* Number of bytes sprintf()ed */
	}
		
	/* else */

	/* Device is a resistor */
	buf += sprintf(buf, ".SUBCKT U%d%d 1 3\n", n1, n2);
	buf += sprintf(buf, "R%d%d 1 3 %u\n", n1, n2, (unsigned int) dev);
	buf += sprintf(buf, "%s", ".ENDS\n");
	*numcomps = 100;	/* Inductor */
	return(buf - tmpp);	/* Number of bytes sprintf()ed */

}


