/* 	$Id: unitor.h,v 1.6 2007/08/31 13:20:15 cgm22 Exp cgm22 $	 */

#define roundup(x, y) ((((x)+((y)-1))/(y))*(y)) /* Roundup to the
						 * nearest multiple of y
						 */

enum {
	MAXNODES = 14,
	MAXDEVS = ((MAXNODES - 1) * MAXNODES / 2),
	MAXTOURNAMENTS = 100000,
	POPULATION = 50,

	NETLISTSZ = (100 * MAXDEVS) + 215, /* buffer size for entire netlist */

	SAMPLEMAX = 27, /* Number of frequency samples. See the .AC line in the netlist. */

	/* XXX: Every SKIP and SHORT value must be unique for rlc.c: verify with the dpool[] LUT: 
	 * This is OK because the values are not directly used or put into netlists.
	 */

	LUTSKIPRES = 1, 
	LUTSHORTRES = 2,

	LUTSKIPCAP = 3, 
	LUTSHORTCAP = 4,

	LUTSKIPIND = 5,
	LUTSHORTIND = 6, 

	MFACTOR = -10 * MAXDEVS

};

#define SPICEBIN "/usr/bin/ngspice"
#define TMPDIR "/var/scratch/cgm22/unitor/"
#define LOGFILEPREFIX "/tmp/unitor-run"


typedef unsigned char allele_t;

extern char logfilename[255];
extern FILE *fd_logfile;

extern int GENELENGTH;
extern int GENELENGTHB;

struct samplesim {
	unsigned int num;	/* Used as index for reading in ngspice dumps,
				 * and weight for tgt[]. Important: This needs
				 * to be unsigned for the float conversions to 
				 * work properly 
				 */ 
	float freq, VM; /* VM is the Voltage Magnitude. Refer to the .AC line */
};

void
ga(char **, float [], int [], char [][42], char *, float (*)(char *, char *));

/* Binary to Gray and vice versa. See functions below for 
 * copyright information.
 */

allele_t binaryToGray8bit(allele_t);
allele_t grayToBinary8bit(allele_t);

allele_t getgeneplusplus(allele_t **, int *, int);

int getbit(char *, int);
int setbit(char *, int, int);

#define CONSOLEDUMP	/* Debug messages to console */

#ifndef _MICROBIAL_GA_
#define _MICROBIAL_GA_
#endif

int voutnodedecode(allele_t);

float lifecycle(char *, char *);
float fitness_correlation(struct samplesim *, int);

int gettimestamp(char *);

void run_simulation(char *, char *);


/* {weight, freq, VM} */

#define		BRW	-1/64U /* Band Reject Weight term */
#define		BPW	-1U	/* Band Pass Weight term */


#define		HP3K	 {{BRW, 0, 0}, {BRW, 1000.0, 0}, {BRW, 2000, 0},	\
			  {BPW, 3000, 1}, {BPW, 4000, 1}, {BPW, 5000, 1},	\
			  {BPW, 6000, 1}, {BPW, 7000, 1}, {BPW, 8000, 1},	\
			  {BPW, 9000, 1}, {BPW, 10000, 1}, {BPW, 11000, 1}	\
			}
				

#define		LP8K	 {{BRW, 0, 1}, {BRW, 1000.0, 1}, {BRW, 2000, 1},	\
			  {BPW, 3000, 1}, {BPW, 4000, 1}, {BPW, 5000, 1},	\
			  {BPW, 6000, 1}, {BPW, 7000, 1}, {BPW, 8000, 1},	\
			  {BRW, 9000, 0}, {BRW, 10000, 0}, {BRW, 11000, 0}	\
			}
				


#define		BP3H8H	 {{BRW, 0, 0}, {BRW, 100.0, 0}, {BRW, 200, 0},		\
			  {BPW, 300, 1}, {BPW, 400, 1}, {BPW, 500, 1},		\
			  {BPW, 600, 1}, {BPW, 700, 1}, {BPW, 800, 1},		\
			  {BRW, 1000, 0}, {BRW, 5000, 0}, {BRW, 11000, 0}	\
			}


#define		BP8K11K	 {{BRW, 0, 0}, {BRW, 1000.0, 0}, {BRW, 2000, 0},	\
			  {BRW, 3000, 0}, {BRW, 4000, 0}, {BRW, 5000, 0},	\
			  {BRW, 6000, 0}, {BPW, 7000, 1}, {BPW, 8000, 1},	\
			  {BPW, 9000, 1}, {BPW, 10000, 1}, {BPW, 11000, 1},	\
			  {BRW, 13000, 0}, {BRW, 14000, 0}, {BRW,15000, 0},	\
			}


#define		BR3K8K	 {{BRW, 0, 1}, {BRW, 1000.0, 1}, {BRW, 2000, 1},		\
			  {BPW, 3000, 0}, {BPW, 4000, 0}, {BPW, 5000, 0},		\
			  {BPW, 6000, 0}, {BPW, 7000, 0}, {BPW, 8000, 0},		\
			  {BRW, 9000, 1}, {BRW, 10000, 1}, {BRW, 11000, 1}		\
			}



#define		MBP3H8H8K11K	 {{BRW, 0, 0}, {BRW, 100.0, 0}, {BRW, 200, 0},		\
			  {BPW, 300, 1}, {BPW, 400, 1}, {BPW, 500, 1},		\
			  {BPW, 600, 1}, {BPW, 700, 1}, {BPW, 800, 1},		\
			  {BRW, 1000, 0}, {BRW, 5000, 0}, {BRW, 5500, 0},	\
			  {BRW, 6000, 0}, {BPW, 7000, 1}, {BPW, 8000, 1},	\
			  {BPW, 9000, 1}, {BPW, 10000, 1}, {BPW, 11000, 1},	\
			  {BRW, 13000, 0}, {BRW, 14000, 0}, {BRW,15000, 0}	\
			}


#define		MBR38K132K135K	 {{BRW, 0, 1}, {BRW, 1000.0, 1}, {BRW, 2000, 1},		\
				  {BPW, 3000, 0}, {BPW, 4000, 0}, {BPW, 5000, 0},		\
				  {BPW, 6000, 0}, {BPW, 7000, 0}, {BPW, 8000, 0},		\
				  {BRW, 9000, 1}, {BRW, 15000, 1}, {BRW, 125000, 1},		\
				  {BPW, 132000, 0}, {BPW, 133000, 0}, {BPW, 135000, 0},		\
				  {BRW, 136000, 1}, {BRW, 138000, 1}, {BRW, 150000, 1}		\
				}


#define		MBR300TO80032K35K	{{BRW, 0, 0}, {BRW, 100.0, 0}, {BRW, 200, 0},		\
					 {BPW, 300, 1}, {BPW, 400, 1}, {BPW, 500, 1},		\
					 {BPW, 600, 1}, {BPW, 700, 1}, {BPW, 800, 1},		\
				  {BRW, 1000, 0}, {BRW, 15000, 0}, {BRW, 25000, 0},		\
				  {BPW, 32000, 1}, {BPW, 33000, 1}, {BPW, 35000, 1},		\
				}
